from typing import ClassVar, List, Union

from attr import attrs, attrib
from statham import converters as con, validators as val
from statham.validators import NotPassed


@attrs(kw_only=True)
class NestedSchema:
    """nested_schema"""

    _required: ClassVar[List[str]] = ["id"]

    id: str = attrib(
        validator=[
            val.instance_of(str),
            val.has_format("uuid"),
            val.pattern(
                "^[0-9a-fA-F]{8}\\-[0-9a-fA-F]{4}\\-[0-9a-fA-F]{4}\\-[0-9a-fA-F"
                "]{4}\\-[0-9a-fA-F]{12}$"
            ),
        ]
    )
    timestamp: Union[str, NotPassed] = attrib(
        validator=[
            val.instance_of(str),
            val.has_format("date-time"),
            val.pattern(
                "^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[1"
                "2][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\\.[0-9]"
                "+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$"
            ),
        ],
        default=NotPassed(),
    )
    version: Union[int, NotPassed] = attrib(
        validator=[val.instance_of(int)], default=0
    )
    annotation: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str)], default="unannotated"
    )


@attrs(kw_only=True)
class SimpleSchema:
    """This is a simple object."""

    _required: ClassVar[List[str]] = ["related"]

    related: NestedSchema = attrib(
        validator=[val.instance_of(NestedSchema)],
        converter=con.instantiate(NestedSchema),  # type: ignore
    )
    amount: Union[float, NotPassed] = attrib(
        validator=[val.instance_of(float)], default=NotPassed()
    )
    children: Union[List[Union[NestedSchema, NotPassed]], NotPassed] = attrib(
        validator=[val.instance_of(list)],
        converter=con.map_instantiate(NestedSchema),  # type: ignore
        default=NotPassed(),
    )



from functools import partial
from logging import getLogger
import re
from typing import Callable, Dict, Type, Union
from uuid import UUID
import warnings

# False positive on `ParserError` import.
from dateutil.parser import parse as parse_datetime, ParserError  # type: ignore

from statham.exceptions import ValidationError


LOGGER = getLogger(__name__)


class _FormatString:
    """Extendable format string register."""

    def __init__(self, name: str):
        self._callable_register: Dict[str, Callable[[str], bool]] = {}
        self.__name__: str = name

    def register(self, format_string: str) -> Callable:
        def _register_callable(is_format: Callable[[str], bool]):
            self._callable_register[format_string] = is_format

        return _register_callable

    def __call__(self, format_string: str, value: str) -> bool:
        if format_string not in self._callable_register:
            full_name = f"{self.__module__}.{self.__name__}"
            warnings.warn(
                (
                    f"No validator found for format string {format_string}. "
                    f"To register new formats, please register a checker with "
                    f"{full_name} as follows:\n"
                    f"```\n"
                    f"@{self.__name__}({format_string})\n"
                    f"def is_{format_string.replace('-', '_')}value) -> bool:\n"
                    f"    ...\n"
                    f"```"
                ),
                RuntimeWarning,
            )
            return True
        return self._callable_register[format_string](value)


format_checker: _FormatString = _FormatString("format_checker")


@format_checker.register("uuid")
def _is_uuid(value: str) -> bool:
    try:
        UUID(value)
    except (ValueError, TypeError):
        return False
    return True


@format_checker.register("date-time")
def _is_date_time(value: str) -> bool:
    try:
        parse_datetime(value)
    except (ParserError, TypeError):
        return False
    return True


def on_types(*type_args: Type) -> Callable:
    """Decorator factory to limit scope of validators to given types."""

    def _validate_if_types(validator: Callable) -> Callable:
        """Return a wrapped version of `validator` to negotiate type."""

        def _inner_validator(instance, attribute, value) -> None:
            if not isinstance(value, type_args):
                return
            validator(instance, attribute, value)

        return _inner_validator

    return _validate_if_types


def raises(message: str) -> Callable:
    """Decorator factory which declares error message for validator."""

    def validate_with_error_message(validator: Callable) -> Callable:
        """Return an attrs compatible validator which raises outer error.

        :param validator: Similar to `attrs` validator interface, but
            accepting an additional argument, which is the error to
            raise if validation fails.
        :return: `attrs` compliant validator.
        """

        def inner_validator(instance, attribute, value) -> None:
            validator(
                instance,
                attribute,
                value,
                partial(ValidationError, instance, attribute, value, message),
            )

        return inner_validator

    return validate_with_error_message


def min_items(minimum_items: int) -> Callable:
    @on_types(list)
    @raises(f"Must contain at least {minimum_items} items.")
    def _min_items(_instance, _attribute, value, error):
        if len(value) < minimum_items:
            raise error()

    return _min_items


def max_items(maximum_items: int) -> Callable:
    @on_types(list)
    @raises(f"Must contain fewer than {maximum_items} items.")
    def _max_items(_instance, _attribute, value, error):
        if len(value) > maximum_items:
            raise error()

    return _max_items


def minimum(minimum_value: Union[int, float]) -> Callable:
    @on_types(int, float)
    @raises(f"Must be greater than or equal to {minimum_value}.")
    def _minimum(_instance, _attribute, value, error):
        if value < minimum_value:
            raise error()

    return _minimum


def maximum(maximum_value: Union[int, float]) -> Callable:
    @on_types(int, float)
    @raises(f"Must be less than or equal to {maximum_value}.")
    def _maximum(_instance, _attribute, value, error):
        if value > maximum_value:
            raise error()

    return _maximum


def exclusive_minimum(exclusive_minimum_value: Union[int, float]) -> Callable:
    @on_types(int, float)
    @raises(f"Must be strictly greater than {exclusive_minimum_value}.")
    def _exclusive_minimum(_instance, _attribute, value, error):
        if value <= exclusive_minimum_value:
            raise error()

    return _exclusive_minimum


def exclusive_maximum(exclusive_maximum_value: Union[int, float]) -> Callable:
    @on_types(int, float)
    @raises(f"Must be strictly less than {exclusive_maximum_value}.")
    def _exclusive_maximum(_instance, _attribute, value, error):
        if value >= exclusive_maximum_value:
            raise error()

    return _exclusive_maximum


def multiple_of(multiple_value: Union[int, float]) -> Callable:
    @on_types(int, float)
    @raises(f"Must be a multiple of {multiple_value}.")
    def _multiple_of(_instance, _attribute, value, error):
        if value % multiple_value:
            raise error()

    return _multiple_of


def has_format(format_string: str) -> Callable:
    @on_types(str)
    @raises(f"Must match format described by {repr(format_string)}.")
    def _format(_instance, _attribute, value, error):
        if not format_checker(format_string, value):
            raise error()

    return _format


def pattern(re_pattern: str) -> Callable:
    @on_types(str)
    @raises(f"Must match regex pattern {repr(re_pattern)}.")
    def _pattern(_instance, _attribute, value, error):
        if not re.match(re_pattern, value):
            raise error()

    return _pattern


def min_length(min_length_value: int) -> Callable:
    @on_types(str)
    @raises(f"Must be at least {min_length_value} characters long.")
    def _min_length_value(_instance, _attribute, value, error):
        if len(value) < min_length_value:
            raise error()

    return _min_length_value


def max_length(max_length_value: int) -> Callable:
    @on_types(str)
    @raises(f"Must be at most {max_length_value} characters long.")
    def _max_length_value(_instance, _attribute, value, error):
        if len(value) > max_length_value:
            raise error()

    return _max_length_value


class NotPassed:
    """Singleton similar to NoneType.

    To distinguish between arguments not passed, and arguments passed as
    None.
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        """Enforce singleton."""
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def __repr__(self) -> str:
        return "NotPassed"

    def __bool__(self) -> bool:
        return False


def instance_of(*types: Type) -> Callable:
    type_names = f"({', '.join((type_.__name__ for type_ in types))})"

    @raises(f"Must be of type {type_names}.")
    def _instance_of(instance, attribute, value, error):
        # Validator acts as a method.
        # pylint: disable=protected-access
        if value == NotPassed() and attribute.name not in instance._required:
            return
        if not isinstance(value, types):
            raise error()

    return _instance_of


SCHEMA_ATTRIBUTE_VALIDATORS: Dict[str, Callable] = {
    "minItems": min_items,
    "maxItems": max_items,
    "minimum": minimum,
    "maximum": maximum,
    "exclusiveMinimum": exclusive_minimum,
    "exclusiveMaximum": exclusive_maximum,
    "multipleOf": multiple_of,
    "format": has_format,
    "pattern": pattern,
    "minLength": min_length,
    "maxLength": max_length,
}
"""Mapping of `Schema` attributes to validators."""



from typing import Iterable, List

from jinja2 import Environment, FileSystemLoader

from statham.constants import NOT_PROVIDED, TypeEnum
from statham.models import ArraySchema, ObjectSchema, Schema
from statham.validators import (
    instance_of,
    NotPassed,
    SCHEMA_ATTRIBUTE_VALIDATORS,
)


def default(schema: Schema, required: bool) -> str:
    default_value = getattr(schema, "default", NOT_PROVIDED)
    if default_value is NOT_PROVIDED:
        if required:
            return ""
        return f"{NotPassed.__name__}()"
    return repr(default_value)


def type_annotation(schema: Schema, required: bool) -> str:
    schema_items = getattr(schema, "items", NOT_PROVIDED)
    mapping = {
        TypeEnum.OBJECT: schema.title,
        TypeEnum.ARRAY: (
            List.__name__
            + (
                f"[{type_annotation(schema_items, False)}]"
                if schema_items is not NOT_PROVIDED
                else ""
            )
        ),
        TypeEnum.INTEGER: int.__name__,
        TypeEnum.NUMBER: float.__name__,
        TypeEnum.STRING: str.__name__,
        TypeEnum.NULL: str(None),
        TypeEnum.BOOLEAN: bool.__name__,
    }
    args = [arg for flag, arg in mapping.items() if flag & schema.type]

    if not required:
        args.append(NotPassed.__name__)
    if len(args) == 1:
        return next(iter(args))
    validator_args = ", ".join(args)
    return f"Union[{validator_args}]"


def validator_type_arg(schema: Schema) -> str:
    mapping = {
        TypeEnum.OBJECT: schema.title,
        TypeEnum.ARRAY: list.__name__,
        TypeEnum.INTEGER: int.__name__,
        TypeEnum.NUMBER: float.__name__,
        TypeEnum.STRING: str.__name__,
        TypeEnum.NULL: "type(None)",
        TypeEnum.BOOLEAN: bool.__name__,
    }
    args = [arg for flag, arg in mapping.items() if flag & schema.type]
    if len(args) == 1:
        return next(iter(args))
    return ", ".join(args)


INDENT = " " * 4


def validators(schema: Schema) -> str:
    return ("\n" + INDENT * 3).join(
        [f"val.{instance_of.__name__}({validator_type_arg(schema)}),"]
        + extra_validators(schema)
    )


def extra_validators(schema: Schema) -> List[str]:
    return [
        (f"val.{validator.__name__}({repr(getattr(schema, attribute))}),")
        for attribute, validator in SCHEMA_ATTRIBUTE_VALIDATORS.items()
        if hasattr(schema, attribute)
        and getattr(schema, attribute) is not NOT_PROVIDED
    ]


def converter(schema: Schema) -> str:
    if isinstance(schema, ObjectSchema):
        return f"con.instantiate({schema.title})"
    if isinstance(schema, ArraySchema) and isinstance(
        schema.items, ObjectSchema
    ):
        return f"con.map_instantiate({schema.items.title})"
    return ""


def serialize_object_schemas(schemas: Iterable[ObjectSchema]) -> str:
    environment = Environment(
        loader=FileSystemLoader("statham/templates"),
        trim_blocks=True,
        lstrip_blocks=True,
    )
    environment.filters["validators"] = validators
    environment.filters["converter"] = converter
    environment.filters["default"] = default
    environment.filters["type_annotation"] = type_annotation
    template = environment.get_template("template.py.j2")
    return template.render(
        schemas=schemas, type_enum=TypeEnum, not_provided=NOT_PROVIDED
    )



from collections import defaultdict
from itertools import chain
import re
from typing import Any, Callable, Dict, Set, Type


# This is a stateful function.
class _Counter:  # pylint: disable=too-few-public-methods
    """Callable which returns the number of calls for the given argument."""

    def __init__(self):
        self.counts = defaultdict(lambda: 0)

    def __call__(self, key: str) -> str:
        self.counts[key] = self.counts[key] + 1
        return f"{key}{self.counts[key]}"


counter: Callable = _Counter()


def dict_map(
    function: Callable[[Any], Any], dictionary: Dict[Any, Any]
) -> Dict[Any, Any]:
    """Similar to `map`, but operates on a dictionary's values."""
    return {key: function(value) for key, value in dictionary.items()}


def dict_filter(
    function: Callable[[Any], bool], dictionary: Dict[Any, Any]
) -> Dict[Any, Any]:
    """Similar to `filter`, but operates on a dictionary's values."""
    return {key: value for key, value in dictionary.items() if function(value)}


def all_subclasses(klass: Type) -> Set[Type]:
    return set(klass.__subclasses__()).union(
        [s for c in klass.__subclasses__() for s in all_subclasses(c)]
    )


def _title_format(string: str) -> str:
    """Convert titles in schemas to class names."""
    words = re.split(r"[ _-]", string)
    segments = chain.from_iterable(
        [
            re.findall("[A-Z][^A-Z]*", word[0].upper() + word[1:])
            for word in words
        ]
    )
    return "".join(segment.title() for segment in segments)



from enum import auto, Flag, unique
from functools import reduce
from typing import Any, Dict, Iterator, List, Tuple, Union


INDENT = " " * 4


JSONElement = Union[Dict[str, Any], List[Any], int, float, bool, None, str]


class NotProvidedType:
    """Singleton for indicating whether a field has been specified.

    Useful to differentiate between e.g. default=None and no default.
    """

    def __repr__(self):
        return "<NOTPROVIDED>"


NOT_PROVIDED: NotProvidedType = NotProvidedType()


IGNORED_SCHEMA_KEYWORDS = (
    # Metadata
    "$id",
    "$schema",
    "definitions",
    "$comment",
    "contentMediaType",
    "contentEncoding",
    "examples",
    "readOnly",
    "writeOnly",
    "deprecated",
    # Not implemented
    "additionalItems",
    "contains",
    "additionalProperties",
    "propertyNames",
    "patternProperties",
    "minProperties",
    "maxProperties",
    "enum",
    "const",
    "uniqueItems",
    # OpenAPI/Swagger
    "example",
    "nullable",
)
"""Keywords to ignore from jsonschemas when parsing."""


@unique
class TypeEnum(Flag):
    OBJECT = auto()
    ARRAY = auto()
    NUMBER = auto()
    INTEGER = auto()
    STRING = auto()
    NULL = auto()
    BOOLEAN = auto()


_JSON_SCHEMA_TYPE_MAP: Tuple[Tuple[str, TypeEnum], ...] = (
    ("object", TypeEnum.OBJECT),
    ("array", TypeEnum.ARRAY),
    ("number", TypeEnum.NUMBER),
    ("integer", TypeEnum.INTEGER),
    ("string", TypeEnum.STRING),
    ("null", TypeEnum.NULL),
    ("boolean", TypeEnum.BOOLEAN),
)

_TYPE_LOOKUP = dict(_JSON_SCHEMA_TYPE_MAP)


def get_flag(*types: str) -> TypeEnum:
    """Get type flag for jsonschema type representation."""
    if not types:
        raise ValueError("Need at least one type.")
    unknown = set(types) - set(_TYPE_LOOKUP)
    if unknown:
        raise ValueError(f"Unknown JSONSchema types: {unknown}")
    if len(types) == 1:
        return _TYPE_LOOKUP[next(iter(types))]
    type_components: Iterator[TypeEnum] = (
        _TYPE_LOOKUP[type_] for type_ in types
    )
    return reduce(lambda x, y: x | y, type_components)


_ENUM_LOOKUP = {enum: keyword for keyword, enum in _JSON_SCHEMA_TYPE_MAP}


def get_type(flag: TypeEnum) -> Union[str, List[str]]:
    """Get jsonschema type representation of type flag."""
    return [_ENUM_LOOKUP[_flag] for _flag in _ENUM_LOOKUP if _flag & flag]



from typing import ClassVar, List, Union

from attr import attrs, attrib
from statham import converters as con, validators as val
from statham.validators import NotPassed


@attrs(kw_only=True)
class Model:
    """Model"""

    _required: ClassVar[List[str]] = []

    number_integer: Union[int, float, NotPassed] = attrib(
        validator=[
            val.instance_of(int, float),
            val.minimum(2.5),
            val.exclusive_maximum(5),
        ],
        default=3.5,
    )
    string_integer: Union[int, str, NotPassed] = attrib(
        validator=[
            val.instance_of(int, str),
            val.minimum(1970),
            val.exclusive_maximum(2050),
            val.has_format("date-time"),
        ],
        default=2000,
    )
    string_null: Union[str, None, NotPassed] = attrib(
        validator=[
            val.instance_of(str, type(None)),
            val.pattern(
                "^[0-9a-fA-F]{8}\\-[0-9a-fA-F]{4}\\-[0-9a-fA-F]{4}\\-[0-9a-fA-F"
                "]{4}\\-[0-9a-fA-F]{12}$"
            ),
        ],
        default=None,
    )



from functools import lru_cache, reduce
from typing import Any, ClassVar, Dict, List, Type, Union

from attr import attrs, attrib, Factory
from attr.validators import instance_of

from statham.constants import (
    get_flag,
    get_type,
    JSONElement,
    NotProvidedType,
    NOT_PROVIDED,
    TypeEnum,
)
from statham.exceptions import SchemaParseError
from statham.helpers import (
    all_subclasses,
    counter,
    dict_map,
    dict_filter,
    _title_format,
)


@attrs(kw_only=True, frozen=True)
class Schema:

    type: ClassVar[TypeEnum]

    _type: Union[str, List[str]] = attrib(validator=[instance_of((str, list))])

    title: str = attrib(validator=[instance_of(str)], converter=_title_format)
    description: str = attrib(
        validator=[instance_of(str)],
        default=Factory(lambda self: self.title, takes_self=True),
    )


def parse_schema(schema: Dict[str, JSONElement]) -> Schema:
    """Convert a JSON Schema schema to a Model Instance.

    Looks up subclasses of Schema and matches on the `type` class
    variable.
    """
    try:
        type_prop = schema["type"]
    except KeyError:
        raise SchemaParseError.missing_type(schema)
    types = type_prop if isinstance(type_prop, list) else [type_prop]
    schema["title"] = schema.get("title") or counter(".".join(types))
    schema["description"] = schema.get("description") or schema["title"]
    try:
        return model_from_types(*types)(**schema)  # type: ignore
    except TypeError as exc:
        raise SchemaParseError.from_exception(schema, exc) from exc


@attrs(kw_only=True, frozen=True)
class ArraySchema(Schema):

    type: ClassVar[TypeEnum] = TypeEnum.ARRAY

    items: Schema = attrib(
        validator=[instance_of(Schema)], converter=parse_schema
    )
    # pylint: disable=invalid-name
    minItems: Union[int, NotProvidedType] = attrib(
        validator=[instance_of((int, NotProvidedType))], default=NOT_PROVIDED
    )
    maxItems: Union[int, NotProvidedType] = attrib(
        validator=[instance_of((int, NotProvidedType))], default=NOT_PROVIDED
    )
    # pylint: enable=invalid-name


def _dict_property_convert(
    dictionary: Dict[str, JSONElement]
) -> Dict[str, Schema]:
    return dict_map(
        parse_schema, dict_filter(lambda val: isinstance(val, dict), dictionary)
    )


@attrs(kw_only=True, frozen=True)
class ObjectSchema(Schema):

    type: ClassVar[TypeEnum] = TypeEnum.OBJECT

    properties: Dict[str, Schema] = attrib(
        validator=[instance_of(dict)], converter=_dict_property_convert
    )
    required: List[str] = attrib(validator=[instance_of(list)], factory=list)


@attrs(kw_only=True, frozen=True)
class PrimitiveSchema(Schema):

    default: Any = attrib(default=NOT_PROVIDED)


@attrs(kw_only=True, frozen=True)
class BooleanSchema(PrimitiveSchema):

    type: ClassVar[TypeEnum] = TypeEnum.BOOLEAN


@attrs(kw_only=True, frozen=True)
class BaseNumericSchema(PrimitiveSchema):

    minimum: Union[int, float, NotProvidedType] = attrib(
        validator=[instance_of((int, float, NotProvidedType))],
        default=NOT_PROVIDED,
    )
    maximum: Union[int, float, NotProvidedType] = attrib(
        validator=[instance_of((int, float, NotProvidedType))],
        default=NOT_PROVIDED,
    )
    # pylint: disable=invalid-name
    exclusiveMinimum: Union[int, float, NotProvidedType] = attrib(
        validator=[instance_of((int, float, NotProvidedType))],
        default=NOT_PROVIDED,
    )
    exclusiveMaximum: Union[int, float, NotProvidedType] = attrib(
        validator=[instance_of((int, float, NotProvidedType))],
        default=NOT_PROVIDED,
    )
    multipleOf: Union[int, float, NotProvidedType] = attrib(
        validator=[instance_of((int, float, NotProvidedType))],
        default=NOT_PROVIDED,
    )
    # pylint: enable=invalid-name


@attrs(kw_only=True, frozen=True)
class IntegerSchema(BaseNumericSchema):

    type: ClassVar[TypeEnum] = TypeEnum.INTEGER


@attrs(kw_only=True, frozen=True)
class NumberSchema(BaseNumericSchema):

    type: ClassVar[TypeEnum] = TypeEnum.NUMBER


@attrs(kw_only=True, frozen=True)
class StringSchema(PrimitiveSchema):

    type: ClassVar[TypeEnum] = TypeEnum.STRING

    format: Union[str, NotProvidedType] = attrib(
        validator=[instance_of((str, NotProvidedType))], default=NOT_PROVIDED
    )
    pattern: Union[str, NotProvidedType] = attrib(
        validator=[instance_of((str, NotProvidedType))], default=NOT_PROVIDED
    )
    # pylint: disable=invalid-name
    minLength: Union[int, NotProvidedType] = attrib(
        validator=[instance_of((int, NotProvidedType))], default=NOT_PROVIDED
    )
    maxLength: Union[int, NotProvidedType] = attrib(
        validator=[instance_of((int, NotProvidedType))], default=NOT_PROVIDED
    )
    # pylint: enable=invalid-name


@attrs(kw_only=True, frozen=True)
class NullSchema(Schema):

    type: ClassVar[TypeEnum] = TypeEnum.NULL


def _union_model(*models: Type[Schema]) -> Type[Schema]:
    """Get a model which represents the union of two types."""
    invalid = {ObjectSchema, ArraySchema, PrimitiveSchema, Schema} & set(models)
    if invalid:
        raise SchemaParseError.unsupported_type_union(invalid, set(models))
    model_type = reduce(lambda x, y: x | y, map(lambda _: _.type, models))
    type_names = sorted([name.title() for name in get_type(model_type)])
    name = "Or".join(type_names) + "Schema"
    attribs = {"type": model_type}
    return attrs(kw_only=True, frozen=True)(
        type(name, tuple(model for model in models), attribs)
    )


@lru_cache(maxsize=None)
def _model_from_types_cached(*types: str) -> Type[Schema]:
    """Get the correct schema model from the schema dictionary."""
    flag = get_flag(*types)
    matching_models = [
        SubSchema
        for SubSchema in all_subclasses(Schema)
        if hasattr(SubSchema, "type")
        # Ensure the Sub Schema is one of the fundamental set.
        and (flag & SubSchema.type) is SubSchema.type
        and SubSchema.type in TypeEnum
    ]
    if not matching_models:
        # This block shouldn't be hit!
        # If every type flag has a corresponding fundamental model,
        # there should always be a match.
        raise RuntimeError(  # pragma: no cover
            "No existing models found construct which can construct "
            "the requested type! Does every declared flag on "
            f"`{TypeEnum.__module__}.{TypeEnum.__name__}` have an "
            "equivalent model declared?"
        )
    if len(matching_models) == 1:
        return next(iter(matching_models))
    return _union_model(*matching_models)


def model_from_types(*types: str) -> Type[Schema]:
    """Helper to get schema model from type strings.

    Wraps a cached helper to ensure unions are declared multiple times.
    """
    return _model_from_types_cached(*sorted(types))



import json
from typing import Dict, Set

from statham.constants import JSONElement


class JSONSchemaObjectError(Exception):
    """Base exception"""


class ValidationError(JSONSchemaObjectError):
    """Validation failure in generated models."""

    def __init__(self, instance, attribute, value, message) -> None:
        super().__init__(
            f"Failed validating `{type(instance).__name__}."
            f"{attribute.name} = {repr(value)}`. {message}"
        )


class SchemaParseError(JSONSchemaObjectError):
    """Failure during parsing of provided JSON Schema input."""

    @classmethod
    def missing_type(cls, schema: Dict[str, JSONElement]) -> "SchemaParseError":
        return cls(f"No type defined in schema: {schema}")

    @classmethod
    def unsupported_type_union(
        cls, invalid: Set, requested: Set
    ) -> "SchemaParseError":
        return cls(
            f"Can't produce a union for these types: {invalid}. "
            f"The following union was requested: {requested}"
        )

    @classmethod
    def no_class_equivalent_schemas(cls) -> "SchemaParseError":
        return cls(
            "Schema document contains no object schemas from which to "
            "derive a class."
        )

    @classmethod
    def unresolvable_declaration(cls) -> "SchemaParseError":
        return cls(
            "Schema document has an unresolvable declaration tree. This "
            "generally occurs due to cyclical references."
        )

    @classmethod
    def from_exception(
        cls, schema: Dict[str, JSONElement], exception: Exception
    ) -> "SchemaParseError":
        return cls(
            "Failed to parse the following schema:\nError: "
            f"{str(exception)}:\nSchema: {json.dumps(schema)}"
        )



from typing import ClassVar, List, Union

from attr import attrs, attrib
from statham import converters as con, validators as val
from statham.validators import NotPassed


@attrs(kw_only=True)
class Category:
    """Category with required name."""

    _required: ClassVar[List[str]] = ["required_name"]

    required_name: str = attrib(validator=[val.instance_of(str)])


@attrs(kw_only=True)
class NestedRemote:
    """Local reference nested in remote reference"""

    _required: ClassVar[List[str]] = []

    name: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str)], default=NotPassed()
    )


@attrs(kw_only=True)
class Remote:
    """Remote ref in sub-directory."""

    _required: ClassVar[List[str]] = []

    name: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str)], default=NotPassed()
    )
    nested: Union[NestedRemote, NotPassed] = attrib(
        validator=[val.instance_of(NestedRemote)],
        converter=con.instantiate(NestedRemote),  # type: ignore
        default=NotPassed(),
    )


@attrs(kw_only=True)
class Model:
    """Model defined with remote refs."""

    _required: ClassVar[List[str]] = []

    filesystem_remote_ref_flat: Union[Category, NotPassed] = attrib(
        validator=[val.instance_of(Category)],
        converter=con.instantiate(Category),  # type: ignore
        default=NotPassed(),
    )
    filesystem_remote_ref_directory: Union[Remote, NotPassed] = attrib(
        validator=[val.instance_of(Remote)],
        converter=con.instantiate(Remote),  # type: ignore
        default=NotPassed(),
    )



from typing import ClassVar, List, Union

from attr import attrs, attrib
from statham import converters as con, validators as val
from statham.validators import NotPassed


@attrs(kw_only=True)
class Model:
    """Model"""

    _required: ClassVar[List[str]] = []

    string_no_validation: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str)], default=NotPassed()
    )
    string_default: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str)], default="foo"
    )
    string_format_uuid: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str), val.has_format("uuid")],
        default=NotPassed(),
    )
    string_format_date_time: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str), val.has_format("date-time")],
        default=NotPassed(),
    )
    string_format_uri: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str), val.has_format("uri")],
        default=NotPassed(),
    )
    string_pattern: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str), val.pattern(r"^(foo|bar).*")],
        default=NotPassed(),
    )
    string_minLength: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str), val.min_length(3)], default=NotPassed()
    )
    string_maxLength: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str), val.max_length(3)], default=NotPassed()
    )
    integer_no_validation: Union[int, NotPassed] = attrib(
        validator=[val.instance_of(int)], default=NotPassed()
    )
    integer_default: Union[int, NotPassed] = attrib(
        validator=[val.instance_of(int)], default=1
    )
    integer_minimum: Union[int, NotPassed] = attrib(
        validator=[val.instance_of(int), val.minimum(3)], default=NotPassed()
    )
    integer_exclusiveMinimum: Union[int, NotPassed] = attrib(
        validator=[val.instance_of(int), val.exclusive_minimum(3)],
        default=NotPassed(),
    )
    integer_maximum: Union[int, NotPassed] = attrib(
        validator=[val.instance_of(int), val.maximum(3)], default=NotPassed()
    )
    integer_exclusiveMaximum: Union[int, NotPassed] = attrib(
        validator=[val.instance_of(int), val.exclusive_maximum(3)],
        default=NotPassed(),
    )
    integer_multipleOf: Union[int, NotPassed] = attrib(
        validator=[val.instance_of(int), val.multiple_of(2)],
        default=NotPassed(),
    )
    number_no_validation: Union[float, NotPassed] = attrib(
        validator=[val.instance_of(float)], default=NotPassed()
    )
    number_default: Union[float, NotPassed] = attrib(
        validator=[val.instance_of(float)], default=1.5
    )
    number_minimum: Union[float, NotPassed] = attrib(
        validator=[val.instance_of(float), val.minimum(2.5)],
        default=NotPassed(),
    )
    number_exclusiveMinimum: Union[float, NotPassed] = attrib(
        validator=[val.instance_of(float), val.exclusive_minimum(2.5)],
        default=NotPassed(),
    )
    number_maximum: Union[float, NotPassed] = attrib(
        validator=[val.instance_of(float), val.maximum(2.5)],
        default=NotPassed(),
    )
    number_exclusiveMaximum: Union[float, NotPassed] = attrib(
        validator=[val.instance_of(float), val.exclusive_maximum(2.5)],
        default=NotPassed(),
    )
    number_multipleOf: Union[float, NotPassed] = attrib(
        validator=[val.instance_of(float), val.multiple_of(2.5)],
        default=NotPassed(),
    )
    boolean_no_validation: Union[bool, NotPassed] = attrib(
        validator=[val.instance_of(bool)], default=NotPassed()
    )
    boolean_default: Union[bool, NotPassed] = attrib(
        validator=[val.instance_of(bool)], default=True
    )
    null_no_validation: Union[None, NotPassed] = attrib(
        validator=[val.instance_of(type(None))], default=NotPassed()
    )
    array_no_validation: Union[List[Union[int, NotPassed]], NotPassed] = attrib(
        validator=[val.instance_of(list)], default=NotPassed()
    )
    array_minItems: Union[List[Union[int, NotPassed]], NotPassed] = attrib(
        validator=[val.instance_of(list), val.min_items(1)], default=NotPassed()
    )
    array_maxItems: Union[List[Union[int, NotPassed]], NotPassed] = attrib(
        validator=[val.instance_of(list), val.max_items(2)], default=NotPassed()
    )



from typing import ClassVar, List, Union

from attr import attrs, attrib
from statham import converters as con, validators as val
from statham.validators import NotPassed


@attrs(kw_only=True)
class Category:
    """Category with required name."""

    _required: ClassVar[List[str]] = ["required_name"]

    required_name: str = attrib(validator=[val.instance_of(str)])


@attrs(kw_only=True)
class Child:
    """Model with name and reference to category."""

    _required: ClassVar[List[str]] = []

    name: Union[str, NotPassed] = attrib(
        validator=[val.instance_of(str)], default=NotPassed()
    )
    category: Union[Category, NotPassed] = attrib(
        validator=[val.instance_of(Category)],
        converter=con.instantiate(Category),  # type: ignore
        default=NotPassed(),
    )


@attrs(kw_only=True)
class Model:
    """Model with references to children and category."""

    _required: ClassVar[List[str]] = []

    children: Union[List[Union[Child, NotPassed]], NotPassed] = attrib(
        validator=[val.instance_of(list)],
        converter=con.map_instantiate(Child),  # type: ignore
        default=NotPassed(),
    )
    category: Union[Category, NotPassed] = attrib(
        validator=[val.instance_of(Category)],
        converter=con.instantiate(Category),  # type: ignore
        default=NotPassed(),
    )



from typing import Type

from statham.validators import NotPassed


def instantiate(model: Type):
    """Converter factory for instantiating a model from dictionary.

    This allow passing either a model instance or instantiating a new
    one from a dictionary.
    """

    def _convert(data):
        if isinstance(data, (model, NotPassed)):
            return data
        return model(**data)

    return _convert


def map_instantiate(model: Type):
    """Converter factory for instantiating a list of models.

    This allows passing a mixed list of instances and dictionaries and
    receiving a list of instances.
    """
    _factory = instantiate(model)

    def _convert(list_data):
        if isinstance(list_data, NotPassed):
            return list_data
        return [_factory(data) for data in list_data]

    return _convert



from typing import Dict, Set

from attr import attrs

from statham.exceptions import SchemaParseError
from statham.models import ArraySchema, ObjectSchema, Schema


# This is a dataclass.
# pylint: disable=too-few-public-methods
@attrs(auto_attribs=True)
class ClassDef:
    schema: ObjectSchema
    depends: Set[str]


def _get_first_class_schema(schema: Schema) -> ObjectSchema:
    if isinstance(schema, ObjectSchema):
        return schema
    if isinstance(schema, ArraySchema):
        return _get_first_class_schema(schema.items)
    raise SchemaParseError.no_class_equivalent_schemas()


class ClassDependencyResolver:
    """Iterator which returns classes in declaration order."""

    def __init__(self, schema: Schema):
        self._class_defs: Dict[str, ClassDef] = {}
        _ = self._extract_schemas(_get_first_class_schema(schema))

    def _extract_schemas(self, schema: ObjectSchema) -> Set[str]:
        if schema.title in self._class_defs:
            return self[schema.title].depends
        deps = {schema.title}
        for nested in schema.properties.values():
            try:
                next_schema: ObjectSchema = _get_first_class_schema(nested)
            except SchemaParseError:
                continue
            deps = deps | self._extract_schemas(next_schema)
        self._add(
            schema.title, ClassDef(schema=schema, depends=deps - {schema.title})
        )
        return deps

    def __getitem__(self, key: str) -> ClassDef:
        return self._class_defs[key]

    def __iter__(self) -> "ClassDependencyResolver":
        return self

    def __next__(self) -> ObjectSchema:
        next_item = self._next_key()
        for value in self._class_defs.values():
            value.depends = value.depends - {next_item}
        return self._class_defs.pop(next_item).schema

    def _add(self, key: str, class_def: ClassDef):
        self._class_defs[key] = class_def

    def _next_key(self) -> str:
        try:
            return next(
                key
                for key, value in self._class_defs.items()
                if not value.depends
            )
        except StopIteration as exc:
            if not self._class_defs:
                raise exc
            raise SchemaParseError.unresolvable_declaration()
